------------------------------------------------------
-- Aysel Karimova                                   --
-- SMALLSAT specialisation project-PCA covariance   --
-- Supervisors: Kjetil Svarstad, Milica Orlandic    --
-- NTNU-EMECS  Dec, 2017                            --
-- Updated for the master thesis, 2018              --
-- Covariance module calculation - top level        --
------------------------------------------------------

--Control path with the instantiation of data path 
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use work.covariance_pkg.all;

entity covariance_optimised is
  generic(
    constant data_width  : positive := 16; --size of one input component
    constant depth_array : positive := 2500; --number of components in one plane, 2500 in our case
    constant planes      : positive := 100; --number of planes, 100 in our case
    constant buffers     : positive := 4 --number of buffers
  );
  port(
    x            : in  std_logic_vector(data_width - 1 downto 0);
    clock, reset : in  std_logic;
    enable       : in  std_logic;
    valid_input  : in  std_logic;
    done         : out std_logic;
    valid        : out std_logic_vector(buffers - 1 downto 0);
    column, row  : out output_row_column(buffers - 1 downto 0);
    y            : out output_t(buffers - 1 downto 0));
end covariance_optimised;

architecture behav of covariance_optimised is

  type state_type is (st_busy, st_idle);
  signal state : state_type;
  
  signal count_values  : natural range 0 to depth_array; --components on each plane
  signal count_rows    : natural range 0 to planes-1; --number of input planes each of which correspond to one row in the cov matrix
  signal valid_driver  : std_logic_vector(buffers - 1 downto 0);
  signal count_columns : natural range 0 to planes-1; --depending on the number of planes

  signal column_temp    : output_row_column(buffers - 1 downto 0);
  signal row_temp       : output_row_column(buffers - 1 downto 0);
  signal enable_driver  : std_logic_vector(buffers - 1 downto 0);
  signal ctrl_driver    : std_logic_vector(buffers - 1 downto 0);
  signal retention_driver: std_logic;

  component cov_datapath
    port(
      x             : in  std_logic_vector(15 downto 0);
      clock, reset  : in  std_logic;
      retention     : in  std_logic;
      ctrl, enable  : in  std_logic;
      y             : out signed(31 downto 0));
  end component;

begin

  Datapath_generator : for i in 0 to buffers - 1 generate
  begin
    datapath_inst : cov_datapath
      port map(
        x         => x,
        clock     => clock,
        ctrl      => ctrl_driver(i),
        enable    => enable_driver(i),
        retention => retention_driver,
        reset     => reset,
        y         => y(i));
  end generate Datapath_generator;
  
  retention_driver <= '1' when (count_values /= 0) else '0';
  valid            <= valid_driver;

  column <= column_temp;
  row    <= row_temp;
  
  --sequential logic
   
  control_path : process(reset, clock)
  begin
    if reset = '1' then
      state         <= st_idle;
      count_values  <= 0;
      count_rows    <= 0;
      count_columns <= 0;
      done          <= '0';
      column_temp   <= (others => 0);
      row_temp      <= (others => 0);
      valid_driver  <= (others => '0');
      enable_driver <= (others => '0');
      ctrl_driver   <= (others => '0');
    elsif clock'event and clock = '1' then
      if enable = '1' then
        case state is 
          when st_busy =>
            count_values <= count_values + 1;
            valid_driver <= (others => '0');
            if (count_values = depth_array - 1) then
              ctrl_driver   <= (ctrl_driver(buffers - 2 downto 0) & '0'); --shift ctrl one left
              enable_driver <= (enable_driver(buffers - 2 downto 0) & '1'); --shift enable one left
              valid_driver  <= enable_driver; -- All enabled data paths in previous plane have now finished computation
              count_values  <= 0;
              -- to output rows and columns to use in bram
              for i in 0 to buffers - 1 loop
                if (enable_driver(i) = '1') then
                  column_temp(i) <= count_columns;
                  row_temp(i) <= count_rows + i;
                else 
                  column_temp(i) <= 0;
                  row_temp(i) <= 0;
                end if;               
              end loop;
              if (count_rows < planes - buffers and count_columns = planes - 1) then
                ctrl_driver   <= std_logic_vector(to_unsigned(1, buffers));
                enable_driver <= std_logic_vector(to_unsigned(1, buffers));
                count_columns <= count_rows + buffers;
                count_rows    <= count_rows + buffers;
              elsif (count_rows = planes-buffers and count_columns = planes - 1) then
                done    <= '1';  
                state   <= st_idle;
              else
                count_columns <= count_columns + 1;
              end if;
            end if;
          when st_idle =>
          --finished! Get ready for next computation
            count_values  <= 0;
            count_rows    <= 0;
            count_columns <= 0;
            enable_driver <= (others => '0');
            ctrl_driver   <= (others => '0');
            valid_driver  <= (others => '0');
            if valid_input = '1' then
              done      <= '0';  
              enable_driver <= std_logic_vector(to_unsigned(1, buffers));
              ctrl_driver   <= std_logic_vector(to_unsigned(1, buffers));
              state <= st_busy;
            else
              state <= st_idle;
            end if;
          
        end case;
      end if;
    end if;
  end process control_path;
end behav;
            