------------------------------------------------------
-- Aysel Karimova                                   --
-- SMALLSAT master thesis-PCA Jacobi                --
-- Supervisors: Kjetil Svarstad, Milica Orlandic    --
-- NTNU-EMECS-2018                                  --
-- Eigenpair decomposition for Jacobi               --
------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
library work;
use work.array_pkg.all;

entity eigenanalysis is
  generic(
    constant matrix_size : positive := 100;
    constant n           : positive := 4
  );
  port( 
     clock        : in std_logic;
     reset        : in std_logic;
     enable       : in std_logic;
     valid_input  : in std_logic;
     evalue_p_in  : in row_column_array;
     evalue_q_in  : in row_column_array;  
     evector_p_in : in row_column_array;
     evector_q_in : in row_column_array;
     cos          : in signed(31 downto 0);
     sin          : in signed(31 downto 0);
     p            : in natural range 0 to matrix_size-1;
     q            : in natural range 0 to matrix_size-1;
     evalue_p_out : out row_column_array;
     evalue_q_out : out row_column_array;
     evector_p_out: out row_column_array;
     evector_q_out: out row_column_array;
     valid_output : out std_logic
  );
end eigenanalysis;

architecture Behavioral of eigenanalysis is

  type state_type is (st_idle, st_busy);
  signal state : state_type;

  --intermediate signals for storing the outputs
  signal evalue_p_temp  : row_column_array;
  signal evalue_q_temp  : row_column_array;
  signal evector_p_temp : row_column_array;
  signal evector_q_temp : row_column_array;
  signal counter        : integer;
begin
  --output assignments
  evalue_p_out  <= evalue_p_temp;
  evalue_q_out  <= evalue_q_temp;
  evector_p_out <= evector_p_temp;
  evector_q_out <= evector_q_temp;
  
  jacobi : process(clock, reset)
  --variable declarations
  variable var1, var2, var3, var4 : signed(47 downto 0);
  variable var5, var6             : signed(79 downto 0);
  variable var7, var8             : signed(63 downto 0);
  variable vara, varb             : signed(47 downto 0);
  
  begin
  
    if reset = '1' then
      evalue_p_temp <= (others => "00000000000000000000000000000000");
      evalue_q_temp <= (others => "00000000000000000000000000000000");
      evector_p_temp<= (others => "00000000000000000000000000000000");
      evector_q_temp<= (others => "00000000000000000000000000000000");
      valid_output  <= '0';
      state         <= st_idle;
      counter       <= 0;
    elsif(clock'event and clock = '1') then 
      if enable = '1' then        
        case state is
          when st_busy =>
            for i in 0 to n-1 loop             
              --update formulas for eigenvector matrix
              var1 := cos(31 downto 16) * evector_p_in(n*counter+i) - sin(31 downto 16) * evector_q_in(n*counter+i);
              evector_p_temp(n*counter+i) <= var1(45 downto 14);
  
              var2 := sin(31 downto 16) * evector_p_in(n*counter+i) + cos(31 downto 16) * evector_q_in(n*counter+i);
              evector_q_temp(n*counter+i) <= var2(45 downto 14);
              --update formulas for eigenvalue matrix
              if (n*counter+i /= p) and (n*counter+i /= q) then
                vara := cos(31 downto 16) * evalue_p_in(n*counter+i) - sin(31 downto 16) * evalue_q_in(n*counter+i);
                evalue_p_temp(n*counter+i) <= vara(45 downto 14);
            
                varb := sin(31 downto 16) * evalue_p_in(n*counter+i) + cos(31 downto 16) * evalue_q_in(n*counter+i);
                evalue_q_temp(n*counter+i) <= varb(45 downto 14);
              end if;
            end loop; 
            
              if (counter < (matrix_size/n)-1) then
                 counter <= counter + 1;
                 state <= st_busy;
              else
        --eigenvalue formulas for App, Aqq and Apq
                var5 := cos(31 downto 16) * cos(31 downto 16) * evalue_p_in(p) - 2 * cos(31 downto 16) * sin(31 downto 16) * evalue_p_in(q) + sin(31 downto 16) * sin(31 downto 16) * evalue_q_in(q);
                evalue_p_temp(p) <= var5(59 downto 28);
              
        --out(q,q) = sin*sin*A(p,p)+2*cos*sin*A(p,q)+cos*cos*A(q,q);
                var6 := sin(31 downto 16) * sin(31 downto 16) * evalue_p_in(p) + 2 * cos(31 downto 16) * sin(31 downto 16) * evalue_p_in(q) + cos(31 downto 16) * cos(31 downto 16) * evalue_q_in(q);
                evalue_q_temp(q) <= var6(59 downto 28);
 
        --out(p,q) = (cos*cos-sin*sin)*A(p,q)+cos*sin*(A(p,p)-A(q,q));
                var7 := (cos(31 downto 16)*cos(31 downto 16) - sin(31 downto 16)*sin(31 downto 16)) * evalue_p_in(q) + cos(31 downto 16)*sin(31 downto 16)*(evalue_p_in(p)-evalue_q_in(q));
                evalue_q_temp(p) <= var7(59 downto 28);
                
        --out(q,p) = (cos*cos-sin*sin)*A(p,q)+cos*sin*(A(p,p)-A(q,q));
                var8 := (cos(31 downto 16)*cos(31 downto 16) - sin(31 downto 16)*sin(31 downto 16)) * evalue_p_in(q) + cos(31 downto 16)*sin(31 downto 16)*(evalue_p_in(p)-evalue_q_in(q));
                evalue_p_temp(q) <= var8(59 downto 28);
                
                counter <= 0;
                valid_output <= '1';
                state <= st_idle;
              end if;
              
           when st_idle =>
            if valid_input = '1' then
              valid_output <= '0';
              state <= st_busy;
            else 
              state <= st_idle;
            end if;
        end case;
      end if;
    end if;    
  end process;
end Behavioral;
