------------------------------------------------------
-- Aysel Karimova                                   --
-- SMALLSAT master thesis-PCA Jacobi                --
-- Supervisors: Kjetil Svarstad, Milica Orlandic    --
-- NTNU-EMECS-2018                                  --
-- Top level module for Jacobi                      --
------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

library work;
use work.array_pkg.all;
use work.cordic.all;

entity top_level_cycjacobi is
  generic(
    constant matrix_size : positive := 100;
    constant tolerance   : integer  := 100 --1000000
  );
  port(
    clock           : in  std_logic;
    reset           : in  std_logic;
    enable          : in  std_logic;
    --input data declarations: eigenvalue
    evalue_p_read   : in  row_column_array;
    evalue_q_read   : in  row_column_array;
    --input data for eigenvectors    
    evector_p_read  : in row_column_array;
    evector_q_read  : in row_column_array;
    --control signal declarations
    valid_input     : in  std_logic;
    iteration_done  : out std_logic;
    computation_done: out std_logic;
    --output data declarations: eigenvalue
    evalue_p_write  : out row_column_array;
    evalue_q_write  : out row_column_array;
    --output data eigenvectors
    evector_p_write : out row_column_array;
    evector_q_write : out row_column_array;
    --old p to calculate the write data
    p_latched       : out natural range 0 to matrix_size - 1;   
    q_latched       : out natural range 0 to matrix_size - 1;
    --output updated indexes for reading operation
    p_new           : out natural range 0 to matrix_size - 1;
    q_new           : out natural range 0 to matrix_size - 1
  );
end top_level_cycjacobi;

architecture Behavioral of top_level_cycjacobi is

  type state_type is (st_error, st_update, st_pq, st_delay, st_idle, st_arctan, st_sincos);
  signal state : state_type;

  --component instantiations 
  component off
    port(
      valid_input   : in  std_logic;
      row_p         : in  row_column_array;
      row_q         : in  row_column_array;
      p             : in  natural range 0 to matrix_size - 1;
      q             : in  natural range 0 to matrix_size - 1;
      clock         : in  std_logic;
      reset         : in  std_logic;
      enable        : in  std_logic;
      error         : out signed(71 downto 0);
      valid_output  : out std_logic 
    );
  end component;

  component cordic_sequential is
    generic(
      SIZE               : positive   := 32; --# Width of operands
      ITERATIONS         : positive   := 32; --# Number of iterations for CORDIC algorithm
      RESET_ACTIVE_LEVEL : std_ulogic := '1' --# Asynch. reset control level
    );
    port(
      --# {{clocks|}}
      Clock        : in  std_ulogic; --# System clock
      Reset        : in  std_ulogic; --# Asynchronous reset

      --# {{control|}}
      Data_valid   : in  std_ulogic; --# Load new input data
      Busy         : out std_ulogic; --# Generating new result
      Result_valid : out std_ulogic; --# Flag when result is valid
      Mode         : in  cordic_mode; --# Rotation or vector mode selection

      --# {{data|}}
      X            : in  signed(SIZE - 1 downto 0); --# X coordinate
      Y            : in  signed(SIZE - 1 downto 0); --# Y coordinate
      Z            : in  signed(SIZE - 1 downto 0); --# Z coordinate (angle in brads)

      X_result     : out signed(SIZE - 1 downto 0); --# X result
      Y_result     : out signed(SIZE - 1 downto 0); --# Y result
      Z_result     : out signed(SIZE - 1 downto 0) --# Z result
    );
  end component;

  component sincos_sequential is
    generic(
      SIZE               : positive   := 32; --# Width of operands
      ITERATIONS         : positive   := 32; --# Number of iterations for CORDIC algorithm
      FRAC_BITS          : positive   := 30; --# Total fractional bits
      MAGNITUDE          : real       := 1.0; --# Scale factor for vector length
      RESET_ACTIVE_LEVEL : std_ulogic := '1' --# Asynch. reset control level
    );
    port(
      --# {{clocks|}}
      Clock        : in  std_ulogic; --# System clock
      Reset        : in  std_ulogic; --# Asynchronous reset

      --# {{control|}}
      Data_valid   : in  std_ulogic; --# Load new input data
      Busy         : out std_ulogic; --# Generating new result
      Result_valid : out std_ulogic; --# Flag when result is valid

      Angle        : in  signed(SIZE - 1 downto 0); --# Angle in brads (2**SIZE brads = 2*pi radians)

      --# {{data|}}
      Sin          : out signed(SIZE - 1 downto 0); --# Sine of Angle
      Cos          : out signed(SIZE - 1 downto 0) --# Cosine of Angle
    );
  end component;

  component choose_pq
    port(
      clock  : in  std_logic;
      reset  : in  std_logic;
      enable : in  std_logic;
      p_out  : out natural range 0 to matrix_size - 1;
      q_out  : out natural range 0 to matrix_size - 1
    );
  end component;

  component eigenanalysis
    port( 
       clock          : in std_logic;
       reset          : in std_logic;
       enable         : in std_logic;
       valid_input    : in std_logic;
       evalue_p_in    : in row_column_array;
       evalue_q_in    : in row_column_array;    
       evector_p_in   : in row_column_array;
       evector_q_in   : in row_column_array;
       cos            : in signed(31 downto 0);
       sin            : in signed(31 downto 0);
       p              : in natural range 0 to matrix_size-1;
       q              : in natural range 0 to matrix_size-1;
       evalue_p_out   : out row_column_array;
       evalue_q_out   : out row_column_array;
       evector_p_out  : out row_column_array;
       evector_q_out  : out row_column_array;
       valid_output   : out std_logic
       );
  end component;
    
  --intermediate signals
  signal error                                  : signed(71 downto 0);
  signal arctan_valid_in, sincos_valid_in       : std_logic;
  signal enable_jac, enable_pq, enable_off      : std_logic;
  signal valid_input_jac, valid_output_jac      : std_logic;
  signal valid_input_off, valid_output_off      : std_logic;
  signal evalue_in_p_temp, evalue_in_q_temp     : row_column_array;
  signal evector_in_p_temp, evector_in_q_temp   : row_column_array;
  signal evalue_out_p_temp, evalue_out_q_temp   : row_column_array;
  signal evector_out_p_temp, evector_out_q_temp : row_column_array;
  signal arctan_valid_out, sincos_valid_out     : std_ulogic;
  signal denominator, numerator                 : signed(31 downto 0);
  signal theta, arctan                          : signed(31 downto 0);
  signal cos_cordic, sin_cordic, cos, sin       : signed(31 downto 0);
  signal X_result, Y_result                     : signed(31 downto 0);
  signal sincos_busy, arctan_busy               : std_ulogic;
  signal p, q                                   : natural range 0 to matrix_size - 1;
  
begin

--module instantiations

  off_instance : off
    port map(
      valid_input   => valid_input_off,
      row_p         => evalue_in_p_temp,
      row_q         => evalue_in_q_temp,
      p             => p,
      q             => q,
      clock         => clock,
      reset         => reset,
      enable        => enable_off,
      error         => error,
      valid_output  => valid_output_off
    );

  cordicvector_instance : cordic_sequential
    generic map(
      SIZE       => 32,
      ITERATIONS => 32
    )
    port map(
      Clock        => clock,
      Reset        => reset,
      Data_valid   => arctan_valid_in,
      Busy         => arctan_busy,
      Result_valid => arctan_valid_out,
      Mode         => cordic_vector,
      X            => denominator,
      Y            => numerator,
      Z            => (others => '0'),
      X_result     => X_result,
      Y_result     => Y_result,
      Z_result     => arctan
    );

  cordicrotate_instance : sincos_sequential
    generic map(
      SIZE       => 32,
      ITERATIONS => 32,
      FRAC_BITS  => 30
    )
    port map(
      Clock        => clock,
      Reset        => reset,
      Data_valid   => sincos_valid_in,
      Busy         => sincos_busy,
      Result_valid => sincos_valid_out,
      Angle        => theta,
      Sin          => sin_cordic,
      Cos          => cos_cordic
    );

  choose_pq_instance : choose_pq
    port map(
      clock  => clock,
      reset  => reset,
      enable => enable_pq,
      p_out  => p,
      q_out  => q
    );

   eigenanalysis_instance : eigenanalysis
    port map(
      clock         => clock,
      reset         => reset,
      enable        => enable_jac,
      valid_input   => valid_input_jac,
      evalue_p_in   => evalue_in_p_temp,
      evalue_q_in   => evalue_in_q_temp,      
      evector_p_in  => evector_in_p_temp,
      evector_q_in  => evector_in_q_temp,
      cos           => cos,
      sin           => sin,
      p             => p,
      q             => q,
      evalue_p_out  => evalue_out_p_temp,
      evalue_q_out  => evalue_out_q_temp,
      evector_p_out => evector_out_p_temp,
      evector_q_out => evector_out_q_temp,
      valid_output  => valid_output_jac
    );

  --combinational logic
        
  p_new <= p;
  q_new <= q;
  
  enable_pq  <= '1' when (state = st_pq) else '0';
  enable_jac <= '1'; -- when (state = st_update) else '0';
  enable_off <= '1';-- when (state = st_error) else '0';
  
  --the inputs for cordic vectoring: the values are the inputs for arctan to calculate the angle 
  --formula is angle = 1/2 * arctan (2*Apq / (Aqq - App))
  denominator   <= (evalue_in_q_temp(q)-evalue_in_p_temp(p));
  numerator     <= (evalue_in_p_temp(q)(30 downto 0) & '0'); --2*A(p,q)
  theta         <= (arctan(31) & arctan(31 downto 1)); -- divide by 2
  
  --sequential logic
  process(clock, reset)
  begin
    if reset = '1' then
      state             <= st_idle;
      arctan_valid_in   <= '0';
      sincos_valid_in   <= '0';
      computation_done  <= '0';
      iteration_done    <= '0';
      valid_input_off   <= '0';
      valid_input_jac   <= '0';
      cos               <= "00000000000000000000000000000000";
      sin               <= "00000000000000000000000000000000";
      evalue_p_write    <= (others => "00000000000000000000000000000000");
      evalue_q_write    <= (others => "00000000000000000000000000000000");
      evector_p_write   <= (others => "00000000000000000000000000000000");
      evector_q_write   <= (others => "00000000000000000000000000000000");
      evalue_in_p_temp  <= (others => "00000000000000000000000000000000");
      evalue_in_q_temp  <= (others => "00000000000000000000000000000000");
      evector_in_p_temp <= (others => "00000000000000000000000000000000");
      evector_in_q_temp <= (others => "00000000000000000000000000000000");
    elsif clock'event and clock = '1' then
      if enable = '1' then
        case state is

          when st_arctan =>   -- calculate cordic arctan
            arctan_valid_in <= '0';
            if (arctan_valid_out = '1') then
              sincos_valid_in <= '1';
              if (sincos_valid_out = '0') then
                valid_input_off <= '1';
                state <= st_error;
              end if;
            end if;

          when st_error =>    -- calculate error function
            valid_input_off <= '0';
            if (valid_output_off = '1') then
                state <= st_sincos;
            end if;

          when st_sincos =>   -- calculate (sin(theta), cos(theta))
            sincos_valid_in <= '0';
            if (sincos_valid_out = '1') then
              valid_input_jac <= '1';
              sin <= sin_cordic;
              cos <= cos_cordic;
              state <= st_update;
            else
              state <= st_sincos;
            end if;

          when st_update =>   -- calculate jacobi formulas
             valid_input_jac <= '0';
             if (valid_output_jac = '1') then
                state <= st_pq;
             end if;

          when st_pq =>       -- choose new p and q and update the input
            iteration_done <= '1';
            if (error < tolerance) then
                computation_done <= '1';
            end if;
            state <= st_delay;

          when st_delay =>    -- delay
            evalue_p_write <= evalue_out_p_temp;
            evalue_q_write <= evalue_out_q_temp;
            evector_p_write <= evector_out_p_temp;
            evector_q_write <= evector_out_q_temp;
            iteration_done <= '0';
            arctan_valid_in <= '1';
            state <= st_idle;

          when st_idle =>
            if valid_input = '1' then
              p_latched <= p;
              q_latched <= q;
              evalue_in_p_temp <= evalue_p_read;
              evalue_in_q_temp <= evalue_q_read;
              evector_in_p_temp <= evector_p_read;
              evector_in_q_temp <= evector_q_read;
              arctan_valid_in <= '1';
              if (arctan_valid_out = '0') then
                computation_done <= '0';
                iteration_done   <= '0';
                state <= st_arctan;
              end if;
            else
              state <= st_idle;
            end if;
        end case;
      end if;
    end if;
  end process;
end Behavioral;
